@@ -60,27 +60,6 @@ typedef struct {
 
 
 static PVector* EMPTY_VECTOR = NULL;
-static PyObject* pmap_factory_fn = NULL;
-static PyObject* set_in_fn_name = NULL;
-
-static PyObject* setInPmap(PyObject* keySequence, Py_ssize_t keySize, PyObject* value) {
-  if(pmap_factory_fn == NULL) {
-    // Lazy import pmap factory to avoid circular import problems
-    pmap_factory_fn = PyObject_GetAttrString(PyImport_ImportModule("pyrsistent"), "pmap");
-  }
-
-  PyObject* emptyMap = PyObject_CallFunctionObjArgs(pmap_factory_fn, NULL);
-  PyObject* newSequence = PySequence_GetSlice(keySequence, 1, keySize);
-  PyObject* newMap = PyObject_CallMethodObjArgs(emptyMap,
-						set_in_fn_name,
-						newSequence,
-						value,
-						NULL);
-  Py_DECREF(emptyMap);
-  Py_DECREF(newSequence);
-  return newMap;
-}
-
 static PyObject* transform_fn = NULL;
 
 static PyObject* transform(PVector* self, PyObject* args) {
@@ -141,6 +120,7 @@ static VNode* copyNode(VNode* source) {
   memcpy(result->items, source->items, sizeof(source->items));
   
   for(i = 0; i < BRANCH_FACTOR; i++) {
+    // TODO-OPT: Any need to go on when the first NULL has been found?
     if(result->items[i] != NULL) {
       INC_NODE_REF_COUNT((VNode*)result->items[i]);
     }
@@ -182,7 +162,7 @@ static VNode* nodeFor(PVector *self, int i){
     return node;
   }
 
-  PyErr_SetString(PyExc_IndexError, "Index out of range");
+  PyErr_Format(PyExc_IndexError, "Index out of range: %i", i);
   return NULL;
 }
 
@@ -257,7 +237,7 @@ static void PVector_dealloc(PVector *self) {
   Py_TRASHCAN_SAFE_END(self);
 }
 
-static PyObject *toList(PVector *self) {
+static PyObject *PVector_toList(PVector *self) {
   Py_ssize_t i;
   PyObject *list = PyList_New(self->count);
   for (i = 0; i < self->count; ++i) {
@@ -271,7 +251,7 @@ static PyObject *toList(PVector *self) {
 
 static PyObject *PVector_repr(PVector *self) {
   // Reuse the list repr code, a bit less efficient but saves some code
-  PyObject *list = toList(self);
+  PyObject *list = PVector_toList(self);
   PyObject *list_repr = PyObject_Repr(list);
   Py_DECREF(list);
 
@@ -496,7 +476,7 @@ static PyObject* PVector_pickle_reduce(PVector *self) {
   PyObject* pvector_fn = PyObject_GetAttrString(module, "pvector");
   Py_DECREF(module);
 
-  PyObject *list = toList(self);
+  PyObject *list = PVector_toList(self);
   PyObject *arg_tuple = PyTuple_New(1);
   PyTuple_SET_ITEM(arg_tuple, 0, list);
 
@@ -550,10 +530,10 @@ static void copyInsert(void** dest, void** src, Py_ssize_t pos, void *obj) {
 
 static PyObject* PVector_append(PVector *self, PyObject *obj);
 
-static PyObject* PVector_set_in(PVector *self, PyObject *obj);
-
 static PyObject* PVector_transform(PVector *self, PyObject *obj);
 
+static PyObject* PVector_set_in(PVector *self, PyObject *obj);
+
 static PyObject* PVector_set(PVector *self, PyObject *obj);
 
 static PyObject* PVector_mset(PVector *self, PyObject *args);
@@ -587,13 +567,14 @@ static PyMethodDef PVector_methods[] = {
 	{"append",      (PyCFunction)PVector_append, METH_O,       "Appends an element"},
 	{"set",         (PyCFunction)PVector_set, METH_VARARGS, "Inserts an element at the specified position"},
 	{"extend",      (PyCFunction)PVector_extend, METH_O|METH_COEXIST, "Extend"},
-	{"set_in",      (PyCFunction)PVector_set_in, METH_VARARGS, "Insert an element in a nested structure"},
         {"transform",   (PyCFunction)PVector_transform, METH_VARARGS, "Apply one or more transformations"},
+        {"set_in",      (PyCFunction)PVector_set_in, METH_VARARGS, "Set in"},
         {"index",       (PyCFunction)PVector_index, METH_VARARGS, "Return first index of value"},
 	{"count",       (PyCFunction)PVector_count, METH_O, "Return number of occurrences of value"},
         {"__reduce__",  (PyCFunction)PVector_pickle_reduce, METH_NOARGS, "Pickle support method"},
         {"evolver",     (PyCFunction)PVector_evolver, METH_NOARGS, "Return new evolver for pvector"},
 	{"mset",        (PyCFunction)PVector_mset, METH_VARARGS, "Inserts multiple elements at the specified positions"},
+        {"tolist",      (PyCFunction)PVector_toList, METH_NOARGS, "Convert to list"},
 	{NULL}
 };
 
@@ -666,6 +647,8 @@ static PVector* emptyNewPvec(void) {
 }
 
 static void incRefs(PyObject **obj) {
+  // TODO-OPT: Would it be OK to exit on first NULL? Should not be any
+  //           non NULLs beyond a NULL.
   int i;
   for(i = 0; i < BRANCH_FACTOR; i++) {
     Py_XINCREF(obj[i]);
@@ -674,6 +657,7 @@ static void incRefs(PyObject **obj) {
 
 
 static PVector* newPvec(unsigned int count, unsigned int shift, VNode *root) {
+  // TODO-OPT: Introduce object cache
   PVector *pvec = PyObject_GC_New(PVector, &PVectorType);
   debug("pymem alloc_copy %x, ref cnt: %u\n", pvec, pvec->ob_refcnt);
   pvec->count = count;
@@ -841,6 +825,7 @@ static PyObject* PVector_extend(PVector *self, PyObject *iterable) {
         return NULL;
     }
     
+    // TODO-OPT: Use special fast iterator if available
     iternext = *Py_TYPE(it)->tp_iternext;
     PyObject *item = iternext(it);
     if (item == NULL) {
@@ -850,6 +835,8 @@ static PyObject* PVector_extend(PVector *self, PyObject *iterable) {
       return (PyObject *)self;
     } else {
       PVector *newVec = copyPVector(self);
+      // TODO-OPT test using special case code here for extension to
+      // avoid recalculating tail length all the time.
       while(item != NULL) {
         extendWithItem(newVec, item);
         item = iternext(it);
@@ -875,6 +862,8 @@ static PyObject* PVector_append(PVector *self, PyObject *obj) {
   if(tail_size < BRANCH_FACTOR) {
     INC_NODE_REF_COUNT(self->root);
     PVector *new_pvec = newPvec(self->count + 1, self->shift, self->root);
+    // TODO-OPT No need to copy more than the current tail length
+    // TODO-OPT No need to incRefs for all elements all the time
     copyInsert(new_pvec->tail->items, self->tail->items, tail_size, obj);
     incRefs((PyObject**)new_pvec->tail->items);
     debug("append(): new_pvec=%p, new_pvec->tail=%p, new_pvec->root=%p\n",
@@ -907,6 +896,8 @@ static PyObject* PVector_append(PVector *self, PyObject *obj) {
 static VNode* doSet(VNode* node, unsigned int level, unsigned int position, PyObject* value) {
   debug("doSet(): level == %i\n", level);
   if(level == 0) {
+    // TODO-OPT: Perhaps an alloc followed by a reset of reference
+    // count is enough here since we overwrite all subnodes below.
     VNode* theNewNode = newNode();
     copyInsert(theNewNode->items, node->items, position & BIT_MASK, value);
     incRefs((PyObject**)theNewNode->items);
@@ -951,71 +942,18 @@ static PyObject* internalSet(PVector *self, Py_ssize_t position, PyObject *argOb
     // TODO Remove this case?
     return PVector_append(self, argObj);
   } else {
-    PyErr_SetString(PyExc_IndexError, "Index out of range");
+    PyErr_Format(PyExc_IndexError, "Index out of range: %zd", position);
     return NULL;
   }
 }
 
-static PyObject* PVector_set_in(PVector *self, PyObject *args) {
-  PyObject *keySequence = NULL;
-  PyObject *value = NULL;
-
-  if(!PyArg_ParseTuple(args, "OO", &keySequence, &value)) {
-    return NULL;
-  }
-
-  Py_ssize_t keySize = PySequence_Size(keySequence);
-  if(keySize == 0) {
-    Py_INCREF(self);
-    return (PyObject*)self;
-  } else {
-    PyObject *index = PySequence_GetItem(keySequence, 0);
-    Py_DECREF(index);
-    Py_ssize_t keyIndex = PyNumber_AsSsize_t(index, NULL);
-    if (keyIndex == -1 && PyErr_Occurred()) {
-      return NULL;
-    }
-
-    if(keySize == 1) {
-      return internalSet(self, keyIndex, value);
-    } else if(keyIndex == self->count) {
-      PyObject* newMap = setInPmap(keySequence, keySize, value);
-      if(newMap == NULL) {
-      	return NULL;
-      }
-
-      PyObject* newVector = PVector_append(self, newMap);
-      Py_DECREF(newMap);
-      return newVector;
-    } else {
-      PyObject* currentItem = PVector_get_item(self, keyIndex);
-      if(currentItem == NULL) {
-	return NULL;
-      }
-
-      // This is slightly scary calling out to Python code. Could
-      // there be concurrency issues if the GIL is released?
-      PyObject* newSequence = PySequence_GetSlice(keySequence, 1, keySize);
-      PyObject* newItem = PyObject_CallMethodObjArgs(currentItem, set_in_fn_name, newSequence, value, NULL);
-      Py_DECREF(currentItem);
-      Py_DECREF(newSequence);
-      if(newItem == NULL) {
-	return NULL;
-      }
-      
-      PyObject* newVector = internalSet(self, keyIndex, newItem);
-      Py_DECREF(newItem);
-      return newVector;
-    }
-  }
-}
-
-
 static PyObject* PVector_transform(PVector *self, PyObject *obj) {
   return transform(self, obj);
 }
 
-
+static PyObject* PVector_set_in(PVector *self, PyObject *obj) {
+  return transform(self, obj);
+}
 
 /*
  Steals a reference to the object that is inserted in the vector.
@@ -1085,19 +1023,6 @@ PyObject* moduleinit(void) {
     return NULL;
   }
 
-  // Register with Sequence to appear as a proper sequence to the outside world
-  PyObject* c = PyObject_GetAttrString(PyImport_ImportModule("collections"), "Sequence");
-
-  if (PyType_Ready(Py_TYPE(c)) < 0) {
-    printf("Failed to initialize Sequence!\n");
-    return NULL;
-  }
-
-  if(c == NULL) {
-    debug("Was NULL!\n");
-  }
-  
-  PyObject_CallMethod(c, "register", "O", &PVectorType);
 
 #if PY_MAJOR_VERSION >= 3
   m = PyModule_Create(&moduledef);
@@ -1120,9 +1045,6 @@ PyObject* moduleinit(void) {
   Py_INCREF(&PVectorType);
   PyModule_AddObject(m, "PVector", (PyObject *)&PVectorType);
 
-
-  set_in_fn_name = PyUnicode_FromString("set_in");
-
   return m;
 }
 
@@ -1244,6 +1166,7 @@ Evolver to make multi updates easier to work with and more efficient.
 static void PVectorEvolver_dealloc(PVectorEvolver *);
 static PyObject *PVectorEvolver_append(PVectorEvolver *, PyObject *);
 static PyObject *PVectorEvolver_extend(PVectorEvolver *, PyObject *);
+static PyObject *PVectorEvolver_set(PVectorEvolver *, PyObject *);
 static PyObject *PVectorEvolver_subscript(PVectorEvolver *, PyObject *);
 static PyObject *PVectorEvolver_persistent(PVectorEvolver *);
 static Py_ssize_t PVectorEvolver_len(PVectorEvolver *);
@@ -1260,6 +1183,7 @@ static PyMappingMethods PVectorEvolver_mapping_methods = {
 static PyMethodDef PVectorEvolver_methods[] = {
 	{"append",      (PyCFunction)PVectorEvolver_append, METH_O,       "Appends an element"},
 	{"extend",      (PyCFunction)PVectorEvolver_extend, METH_O|METH_COEXIST, "Extend"},
+	{"set",         (PyCFunction)PVectorEvolver_set, METH_VARARGS, "Set item"},
         {"persistent",  (PyCFunction)PVectorEvolver_persistent, METH_NOARGS, "Create PVector from evolver"},
         {"is_dirty",    (PyCFunction)PVectorEvolver_is_dirty, METH_NOARGS, "Check if evolver contains modifications"},
         {NULL,              NULL}           /* sentinel */
@@ -1361,14 +1285,22 @@ static void PVectorEvolver_dealloc(PVectorEvolver *self) {
 
 static PyObject *PVectorEvolver_append(PVectorEvolver *self, PyObject *args) {
   if (PyList_Append(self->appendList, args) == 0) {
-    Py_RETURN_NONE;
+    Py_INCREF(self);
+    return (PyObject*)self;
   }
 
   return NULL;
 }
 
 static PyObject *PVectorEvolver_extend(PVectorEvolver *self, PyObject *args) {
-  return _PyList_Extend((PyListObject *)self->appendList, args);
+  PyObject *retVal = _PyList_Extend((PyListObject *)self->appendList, args);
+  if (retVal == NULL) {
+    return NULL;
+  }
+
+  Py_DECREF(retVal);
+  Py_INCREF(self);
+  return (PyObject*)self;
 }
 
 static PyObject *PVectorEvolver_subscript(PVectorEvolver *self, PyObject *item) {
@@ -1436,6 +1368,27 @@ static VNode* doSetWithDirty(VNode* node, unsigned int level, unsigned int posit
   return resultNode;
 }
 
+/*
+ Steals a reference to the object that is inserted in the vector.
+*/
+static PyObject *PVectorEvolver_set(PVectorEvolver *self, PyObject *args) {
+  PyObject *argObj = NULL;  /* argument to insert */
+  PyObject *position = NULL;
+
+  /* The n parses for size, the O parses for a Python object */
+  if(!PyArg_ParseTuple(args, "OO", &position, &argObj)) {
+    return NULL;
+  }
+
+  if(PVectorEvolver_set_item(self, position, argObj) < 0) {
+    return NULL;
+  }
+
+  Py_INCREF(self);
+  return (PyObject*)self;
+}
+
+
 static int PVectorEvolver_set_item(PVectorEvolver *self, PyObject* item, PyObject* value) {
   if (PyIndex_Check(item)) {
     Py_ssize_t position = PyNumber_AsSsize_t(item, PyExc_IndexError);
@@ -1469,7 +1422,7 @@ static int PVectorEvolver_set_item(PVectorEvolver *self, PyObject* item, PyObjec
     } else if((0 <= position) && (position < (self->newVector->count + PyList_GET_SIZE(self->appendList) + 1))) {
       return PyList_Append(self->appendList, value); 
     } else {
-      PyErr_SetString(PyExc_IndexError, "Index out of range");
+      PyErr_Format(PyExc_IndexError, "Index out of range: %zd", position);
     }
   } else {
     PyErr_Format(PyExc_TypeError, "Indices must be integers, not %.200s", item->ob_type->tp_name);
